from typing import Any


class Error(Exception):
    """General error exception to be extended

    Attributes:
        message -- explanation of the error
        value -- value to be printed
    """

    def __init__(self, value: Any, message: str) -> None:
        self.message = message
        self.value = value

        super().__init__(self.message)

    def __str__(self) -> str:
        return f"{self.value} -> {self.message}"


class KassResponseTimeoutError(Error):
    pass


class KassResponseDataError(Error):
    pass


class KassMissingTokenError(Error):
    pass



from typing import TypedDict, Union, Any, Dict, Tuple, Literal, Type, Optional
import logging
import hmac
from json import JSONDecodeError
import hashlib
from abc import ABC, abstractmethod
import requests
from requests.exceptions import Timeout
from requests.auth import HTTPBasicAuth
from kass_flow.exceptions import (
    KassResponseDataError,
    KassResponseTimeoutError,
    KassMissingTokenError,
)
from kass_flow.interfaces import (
    KassCallbackPaymentDict,
    KassPaymentResponseDict,
    KassRequestPaymentDict,
    KassBillingPaymentResults,
)

logger = logging.getLogger("kass")


class KassBillingBase(ABC):
    def __init__(
        self,
        kass_token: Optional[str] = None,
        kass_url: str = "https://api.kass.is/v1/payments",
    ):
        if not kass_token:
            raise KassMissingTokenError(
                kass_token, f"Kass token cannot be {kass_token}"
            )
        self.kass_token: str = kass_token
        self.kass_url: str = kass_url
        self.kass_request_timeout: int = 5
        self._payment_token: Optional[str] = None

    @property
    def token(self) -> str:
        return self._payment_token

    def create_signature(self, data: KassCallbackPaymentDict) -> str:

        msg = "{}&{}&{}&{}&{}&{}".format(
            data["payment_id"],
            data["transaction_id"],
            data["order"],
            data["amount"],
            data["status"],
            data["completed"],
        )
        signature = hmac.new(
            bytes(self.kass_token, "utf-8"),
            msg=bytes(msg, "utf-8"),
            digestmod=hashlib.sha256,
        ).hexdigest()
        return signature

    def is_signature_valid(self, res: KassCallbackPaymentDict) -> bool:
        signature = self.create_signature(res)
        return signature == res["signature"]

    def _send_payment_request(
        self, payload: KassRequestPaymentDict
    ) -> Tuple[KassBillingPaymentResults, bool]:
        try:
            res: KassPaymentResponseDict = requests.post(
                self.kass_url,
                json=payload,
                auth=HTTPBasicAuth(self.kass_token, ""),
                timeout=self.kass_request_timeout,
            ).json()
        except Timeout:
            error_msg = f"{self.kass_url} did not respond within {self.kass_request_timeout} seconds."
            raise KassResponseTimeoutError(payload, error_msg)
        except JSONDecodeError:
            error_msg = f"Could not parse Kass response: {res}"
            raise KassResponseDataError(payload, error_msg)

        logger.debug(res)
        is_success = res.get("success") == True
        return {"received": res, "submitted": payload}, is_success

    @abstractmethod
    def create_payment_token(self, payload: KassRequestPaymentDict) -> str:
        return ""

    def dispatch(
        self, payload: KassRequestPaymentDict
    ) -> Tuple[KassBillingPaymentResults, bool]:
        payment_token = self.create_payment_token(payload)
        invoice_data: KassRequestPaymentDict = {
            "order": payload["order"],
            "terminal": payload["terminal"],
            "amount": payload["amount"],
            "description": payload["description"],
            "image_url": payload["image_url"],
            "recipient": payload["recipient"],
            "expires_in": payload["expires_in"],
            "notify_url": f"{payload['expires_in']}/{payload['order']}/{payment_token}/",
        }
        result, is_success = self._send_payment_request(invoice_data)
        return result, is_success

    def _sign(self, **kwargs: Union[str, int]) -> str:
        msg = "&".join([str(v) for v in kwargs.values()])

        signature = hmac.new(
            bytes(self.kass_token, "utf-8"),
            msg=bytes(msg, "utf-8"),
            digestmod=hashlib.sha256,
        ).hexdigest()
        return signature


class KassBilling(KassBillingBase):
    def create_payment_token(self, payload: KassRequestPaymentDict) -> str:
        if not self._payment_token:
            self._payment_token = self._sign(
                order_no=payload["order"],
                mobile=payload["recipient"],
                total=payload["amount"],
            )
        return self._payment_token



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kass_flow']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.24.0,<3.0.0']

setup_kwargs = {
    'name': 'kass-flow',
    'version': '0.2.5',
    'description': 'Library to manage payments using the KASS API',
    'long_description': '\nKass Flow\n=========\n\nHelper for the `KASS <https://www.kass.is/>`_ payment gateway. Written for python 3.8.5 (probably works on versions >= 3.6).\n\nSee the `KASS API docs <https://kass.github.io/api/>`_ for more info.\n\nThis module is not affiliated with KASS.\n\nTODO\n----\n\n\n* [x] Create payment\n* [ ] Retreive payment info\n* [ ] Retreive payment status\n* [ ] Cancel payment\n* [ ] Add concurrency with RQ when dispatching multiple payments\n\nUsage\n-----\n\nHere is a short guide to get you started.\n\nThis will dispatch a payment to the KASS API.\n\n.. code-block:: sh\n\n   poetry add kass-flow\n   # or\n   pip install kass-flow\n\n.. code-block:: python\n\n   from kass_flow.kass import KassBilling\n   from kass_flow.interfaces import KassRequestPaymentDict\n\n   kass_token: str = "some-token"\n   kass_url: str = "https://api.kass.is/v1/payments"\n   instance = KassBilling(kass_token, kass_url)\n\n   payload: KassRequestPaymentDict = {\n       "amount": 2199,\n       "description": "Kass bolur",\n       "image_url": "https://photos.kassapi.is/kass/kass-bolur.jpg",\n       "order": "ABC123",\n       "recipient": "7798217",\n       "terminal": 1,\n       "expires_in": 90,\n       "notify_url": "https://example.com/callbacks/kass",\n   }\n\n   result, is_valid = instance.dispatch(payload)\n\nWhen the recipient cancels or pays the requested order KASS will send a callback to the ``notify_url``\\ , which you need to catch on your server and probably validate the signature that is included in the POST payload.\n\nAt some point in time the user will pay/reject the requested order. Here is an example on how to validate the payload.\n\n.. code-block:: python\n\n   from kass_flow.kass import KassBilling\n   from kass_flow.interfaces import KassRequestPaymentDict\n\n   def some_view_that_handles_kass_callback(request):\n       payload = request.data\n       kass_token: str = "some-token"\n       kass_url: str = "https://api.kass.is/v1/payments"\n       instance = KassBilling(kass_token, kass_url)\n\n       if instance.is_signature_valid(payload):\n           # process the payload.\n\nDevelopment\n-----------\n\n.. code-block:: sh\n\n   pip install poetry\n    # to manage envs yourself\n   poetry config virtualenvs.create false\n   poetry install\n   pytest tests\n\nIf you are using VSCode for development there is a ``.vscode/settings.example.json`` for sensible defaults. Since mypy is used you need to install the ``mypyls`` language server.\n\n.. code-block::\n\n   poetry install "https://github.com/matangover/mypyls/archive/master.zip#egg=mypyls[default-mypy]"\n   # or\n   pip install "https://github.com/matangover/mypyls/archive/master.zip#egg=mypyls[default-mypy]"\n',
    'author': 'Jón Levy',
    'author_email': 'nonni@nonni.cc',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/busla/kass-flow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<=3.8',
}


setup(**setup_kwargs)
